#include <stdio.h>
#include <iostream.h>
#include <string.h>
#include <time.h>
#include "Tree.h"

//NOTE: rand() only ranges from 0 to RAND_MAX,
//so this should not exceed RAND_MAX who's value is 0x7fff.
#define TREE_INTEGRITY_TEST_SIZE          30000

//Person struct
class CPerson
{
public:
  char name[80];
  int age;
};

//This derived class provides the operators to sort by age
class CAgeComp : public CPerson
{
public:
  bool operator<(CAgeComp &c) { return age<c.age; }
  bool operator==(int n) { return age==n; }
  bool operator<(int n) { return age<n; }
};

//This derived class provides the operators to sort by name
class CNameComp : public CPerson
{
public:
  bool operator<(CNameComp &c) { return strcmp(name, c.name)<0; }
  bool operator==(char * c) { return strcmp(name, c)==0; }
  bool operator<(char * c) { return strcmp(name, c)<0; }
};


//Callback func to print a hobbit (for age tree)
void WINAPI PrintAgeFunc(CAgeComp * pCharacter, void * pData)
{
  cout << pCharacter->name << "\t" << pCharacter->age << endl;
}

//Callback func to print a hobbit (for name tree)
void WINAPI PrintNameFunc(CNameComp * pCharacter, void * pData)
{
  cout << pCharacter->name << "\t" << pCharacter->age << endl;
}

//Return a filled CPerson struct based on the params
CPerson * MakeChar(char * name, int age)
{
  CPerson * c=new CPerson;
  c->age=age;
  strcpy(c->name, name);
  return c;
}


//integrity test callback
void WINAPI IntCountFunc(int * n, void * pData)
{
  int * p=(int *)pData;
  (*p)++;
}


void main()
{
  //this tree sorts by age. it is also responsible
  //for deleting the items added.
  //the int in the template declaration allows me to sort
  //by int instead of filling another CAgeComp class to search
  CTree<CAgeComp, int> HobbitAgeTree(TRUE);
  
  //this tree sorts by name using the same Item pointers as
  //the Age sorted tree
  //char * to search by string..
  CTree<CNameComp, char *> HobbitNameTree(FALSE);

  CPerson * c;

  //make 5 people and add them to be sorted by both trees
  c=MakeChar("Frodo", 33);
  HobbitAgeTree.Insert((CAgeComp *)c);
  HobbitNameTree.Insert((CNameComp *)c);

  c=MakeChar("Bilbo", 111);
  HobbitAgeTree.Insert((CAgeComp *)c);
  HobbitNameTree.Insert((CNameComp *)c);
  
  c=MakeChar("Pippin", 26);
  HobbitAgeTree.Insert((CAgeComp *)c);
  HobbitNameTree.Insert((CNameComp *)c);

  c=MakeChar("Sam", 27);
  HobbitAgeTree.Insert((CAgeComp *)c);
  HobbitNameTree.Insert((CNameComp *)c);

  c=MakeChar("Merry", 25);
  HobbitAgeTree.Insert((CAgeComp *)c);
  HobbitNameTree.Insert((CNameComp *)c);

  cout << "Printing the initial Hobbit trees:" << endl << endl;
  cout << "Printing Hobbits sorted by ages:" << endl;
  HobbitAgeTree.InFixCallback(PrintAgeFunc, NULL);

  cout << endl << "Printing Hobbits sorted by name:" << endl;
  HobbitNameTree.InFixCallback(PrintNameFunc, NULL);

  //now when Bilbo's name is changed in one tree, it is changed in
  //the other as well
  cout << endl << endl << "Removing Bilbo from the name tree and \
    changing his name to Gumbo." << endl;
  c=HobbitNameTree.Delete("Bilbo");
  strcpy(c->name, "Gumbo");
  cout << "Reinserting \"Gumbo\" to the name tree." << endl;
  HobbitNameTree.Insert((CNameComp *)c);

  cout << endl << endl << "Printing the modified Hobbit trees:"
    << endl << endl;
  cout << "Printing Hobbits sorted by ages:" << endl;
  HobbitAgeTree.InFixCallback(PrintAgeFunc, NULL);

  cout << endl << "Printing Hobbits sorted by name:" << endl;
  HobbitNameTree.InFixCallback(PrintNameFunc, NULL);

/*------------------------------------------------------------

  The below is just there to test the tree code and integrity

------------------------------------------------------------*/

  //Insert TREE_INTEGRITY_TEST_SIZE random elements between 0 and
  //TREE_INTEGRITY_TEST_SIZE-1.
  //If the number already exists, the collision counter
  //is incremented.
  //Then delete 1000 using DeleteMin, delete 1000 using DeleteMax
  //and (try) Deleting 1000 using Delete (log how many were actually deleted)
  //At the end, the the number of collisions + tree item count + number
  //deleted should total TREE_INTEGRITY_TEST_SIZE

  srand((unsigned)time(NULL));
  //used by callback to count tree elements
  int gettotal=0;               
  ///used by callback to count how many times items were found (collisions)
  int findcount=0;              
  int numdeleted=2000;          

  CTree<int> nums(TRUE);

  int * n;
  for (int i=0;i<TREE_INTEGRITY_TEST_SIZE;i++)
  {
    n=new int;
    *n=rand()%TREE_INTEGRITY_TEST_SIZE;

    if (nums.Find(*n))
    {
      delete n;
      findcount++;
    }
    else
      nums.Insert(n);
  }

  //Make sure the tree has atleast 3000 elements to delete
  //otherwise the following will result in an access violation.
  //Probability states it will though...
  //dont rely on nums.ItemCount() to get a total..
  //that will always return the correct total number,
  //instead calculate it using a tree traverse callback.
  nums.InFixCallback(IntCountFunc, &gettotal);
  if (gettotal<3000)
  {
    cout << "The impossible has happenned..." << endl;
    return;
  }

  //Try to delete 3000 elements. Log the number actually deleted by
  //nums.Delete(...)
  gettotal=0;
  for (i=0;i<1000;i++)
  {
    delete nums.DeleteMin();
    delete nums.DeleteMax();
    n=nums.Delete(rand()%TREE_INTEGRITY_TEST_SIZE);
    if (NULL!=n)
    {
      numdeleted++;
      delete n;
    }
  }

  nums.InFixCallback(IntCountFunc, &gettotal);

  if (findcount+gettotal+numdeleted!=TREE_INTEGRITY_TEST_SIZE)
    cout << endl << "Tree integrity test failed." << endl;
  else
    cout << endl << "Tree integrity test passed." << endl;

  cout << endl << "Press any key to exit." << endl;
  getchar(); // pause
}
